/*        Submission to the Fourth ICFP Programming Contest              *
 *                       Lightning Division                              *

 *     Copyright (C) 2001 Christian Sievers (c.sievers@tu-bs.de)         *

 * Ask, and I'll allow you to do almost anything with it.                *
 * For now, you may compile, execute, evaluate and study this program    *
 * AT YOUR OWN RISK.                                                     *

 * This program is neither cool nor hot, and so is its name: ncnh.       *

 * See http://cristal.inria.fr/ICFP2001/prog-contest/task.html for a     *
 * specification of the problem this program tries to solve.             *

 * See ../README for more information about the program.                 */


#include <stdio.h>

#define SPC 0x20
#define CR  0x0D
#define LF  0x0A
#define TAB 0x09

typedef short bool;
typedef enum {B='B', EM='E', I='I', PL='P', S='S', TT='T', U='U',
              SIZE='s', COLOR='c', CLOSE='/', CHAR='.', WS=' '} Elem;

unsigned char val;
bool lastwasSPC;

#define SOMEOTHER '*' /* not complemented to any tag */
#define MAXBUFF 4096
unsigned char buff[MAXBUFF+1] = {SOMEOTHER};
int buffp=0; /* last elem */
int buffs=1; /* first elem */

#define CLOSEBIT 0x80

void outputelem(char e){
  switch (e){
  case EM: fputs("EM",stdout); break;
  case PL: fputs("PL",stdout); break;
  case TT: fputs("TT",stdout); break;
  default: putchar(e);
  }
}

void flushbuffto(int l){
  int i;
  Elem e;
  for(i=buffs;i<=l;i++){
    putchar('<');
    if ((e=buff[i]) & CLOSEBIT)
      putchar('/');
    outputelem(e & ~CLOSEBIT);
    putchar('>');
  }
  if (l==buffp){
    buffp=0;
    buffs=1;
  }else{
    buff[l]=SOMEOTHER;
    buffs=l+1;
  }
}


void flushbuff(void){
  flushbuffto(buffp);
}

void flushbuffforWS(short u){
  int last;
  bool out=0;
  for (last=buffp;(last>=buffs) && (!out) ;last--){
    switch(buff[last] & ~CLOSEBIT){
    case 'r': case 'g': case 'b': case 'c':
    case 'm': case 'y': case 'k': case 'w': if (u) out=1; break;

    case S: case EM: case I: case B: break;

    default: out=1;
    }
  }   
  if (out) last++;
  flushbuffto(last);
}

bool complemented(unsigned char a, unsigned char b){
  return !(a^b^CLOSEBIT);
}

void output(Elem e){
  if (complemented(buff[buffp],e)){
    buffp--;
  }else{
    buff[++buffp]=e;
    if (buffp==MAXBUFF){
      flushbuff();   /* Emergency! */
    }
  }
}

void outputclose(Elem e){
  output(e|CLOSEBIT);
}

void outchar(void){
  flushbuff();
  putchar(val);
  lastwasSPC=0;
}

void outws(bool tt, short u, char size, char color){
  static short lastu;              /* space context for tt=0 */
  static char lastsize, lastcolor; /*                        */
  if (tt){
    flushbuffforWS(u);
    putchar(val);
    lastwasSPC=0; /* well, maybe it was a SPC, but we don't care in TT-mode */
  }else{
    if (!(lastwasSPC && (u==lastu) && (size==lastsize) && 
          ((u==0) || (color==lastcolor)) )){
      flushbuffforWS(u);
      putchar(SPC);
      lastwasSPC=1;
      lastu=u;
      lastsize=size;
      lastcolor=color;
    }
  }
}


Elem nexttok(void){
#define tokis(t) e=t; break

  int c;
  Elem e;
  if ((val=c=getchar()) == '<'){
    switch(e=val=getchar()){
    case 'B':
    case 'E':
    case 'I':
    case 'P':
    case 'S':
    case 'T':
    case 'U':
    case '/': break;
  
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9': tokis(SIZE);
      
    default: /* must be a color then */ tokis(COLOR);
    }

    /* read rest of tag */
    while (getchar() != '>')
      ;

  }else{

    switch(c){
    case EOF: flushbuff(); exit(0);
    case SPC: case CR: case LF: case TAB: tokis(WS);
    default: /* must be a regular char then */ tokis(CHAR);
    }
  }
  return e;
}


#define parsemore parse(b,em,i,s,tt,u,size,color)
#define parsemorein(e,b,em,i,s,tt,u,size,color) \
        ch=e;output(ch);parse(b,em,i,s,tt,u,size,color);outputclose(ch)

void parse(bool b, bool em, bool i, bool s, bool tt,
           short u, short size, char color){
  char ch;
  for (;;){
    switch (nexttok()){
    case B:
      if (b){
	parsemore;
      }else{
	parsemorein(B,1,em,i,s,tt,u,size,color);
      }
      break;
    case EM:
      if (s){
	parsemore;
      }else{
	parsemorein(EM,b,!em,i,s,tt,u,size,color);
      }
      break;
    case I:
      if (i){
	parsemore;
      }else{
	parsemorein(I,b,em,1,s,tt,u,size,color);
      }
      break;
    case PL:
      if (u||s||em||i||b||tt){
	parsemorein(PL,0,0,0,0,0,0,size,color);
      }else{
	parsemore;
      }
      break;
    case S:
      if (s){
	parsemore;
      }else{
	parsemorein(S,b,0,i,1,tt,u,size,color);
      }
      break;
    case TT:
      if (tt){
	parsemore;
      }else{
	parsemorein(TT,b,em,i,s,1,u,size,color);
      }
      break;
    case U:
      if (u==3){
	parsemore;
      }else{
	parsemorein(U,b,em,i,s,tt,u+1,size,color);
      }
      break;
    case SIZE:
      if (val==size){
	parsemore;
      }else{
	parsemorein(val,b,em,i,s,tt,u,val,color);
      }
      break;
    case COLOR:
      if (val==color){
	parsemore;
      }else{
	parsemorein(val,b,em,i,s,tt,u,size,val);
      }
      break;
    case CHAR:
      outchar();
      break;
    case WS:
      outws(tt,u,size,color);
      break;
    case CLOSE:
      return;
    }
  }
}

int main(void){
  parse(0,0,0,0,0,0,'*','*');
  return 0;
}
